__author__ = 'nickmab'

import types
import threadpool
from . import _utils
from _utils import validate_type as _validate_type

'''Contains convenience classes for working with basic ThreadPools and similar basic tasks.'''

class FuncWrap:
    '''Class for wrapping a function call, including remembering
    its args, return value (or exception) and optionally naming it. Useful in conjunction
    with basic ThreadPools where you do not wish to implement
    a callback function to handle new results.'''
    def __init__(self, target=None, name='', args=(), kwargs={}):
        '''takes a target func, args (tup) and kwargs (dict) for that func and 
        optional name (str) (for identification) to run later by calling FuncWrap.run'''
        # validation first
        if not target:
            raise ValueError('Must provide a valid target function at a minimum.')
        _validate_type('target', target, types.FunctionType)
        _validate_type('name', name, types.StringType)
        _validate_type('args', args, types.TupleType)
        _validate_type('kwargs', kwargs, types.DictionaryType)

        self.target = target 
        self.name = name
        self.args = args
        self.kwargs = kwargs
        self.result = None
        self.exception = None

    def run(self):
        '''call the target function with the specified args / kwargs within a 
        try/except and store the result (or exception).'''
        try:
            self.result = self.target(*self.args, **self.kwargs)
        except Exception as e:
            self.exception = e
        

class FuncWrapPool:
    '''Class for creating a ThreadPool with a specified number of workers and 
    running through an iterable of FuncWrap instances conveniently.'''
    def __init__(self, num_workers=5, func_wrap_instances_iterable=[]):
        '''Specify a number of worker threads and provide an iterable of FuncWrap instances'''
        # validation and logic inline
        if not func_wrap_instances_iterable:
            raise ValueError('Must provide a non-empty iterable with FuncWrap instances!')
        _validate_type('num_workers', num_workers, types.IntType)
        self.num_workers = num_workers
        self.func_wraps = []
        for fw in func_wrap_instances_iterable:
            try:
                _validate_type('func_wrap', fw, FuncWrap)
            except TypeError as e:
                raise TypeError('FuncWrapPool expects an iterable of FuncWrap instances. ' +
                    'Instead got %s as part of %s.' %(fw, func_wrap_instances_iterable))
            self.func_wraps.append(fw)
        
    def run(self):
        '''start a ThreadPool with (self.num_workers) workers, run all our FuncWrap instances
        and wait for the results to all come in.'''
        pool = threadpool.ThreadPool(self.num_workers)
        [pool.putRequest(threadpool.WorkRequest(fw.run)) for fw in self.func_wraps]
        pool.wait()



__author__ = 'nickmab'

from setuptools import setup, find_packages

setup(
    name='nickmab.async_util',
    version='0.1',
    license='MIT',
    author='Nick Mabbutt',
    author_email='nickmab@gmail.com',
    description='My first package (be nice please!). Convenience objects for working with long-running calcs and web requests.',
    url='https://github.com/nickmab/async_util',
    keywords='json thread async',
    packages=find_packages(),
    namespace_packages=['nickmab'],
    install_requires=[
        'beautifulsoup4>=4.3.2',
        'threadpool>=1.2.7',
    ],
)



__author__ = 'nickmab'

'''Module for doing web requests with the help of nickmab.async_util.threads module.'''

import types
import urllib2
import json
from . import threads as thr
from _utils import validate_type as _validate_type

class JSONQueryPool:
    '''Class for grouping together many web requests which should return JSON data,
    running them asynchronously, storing the results and automatically parsing to py objs.'''
    def __init__(self, num_worker_threads=5, queries={}):
        '''Takes a dict of { 'query_str_label': 'query_url' } which should consist of
        queries that return a JSON object which will later be retrieved and parsed.'''
        # first define a function that we're going to use to fetch url results.
        def fetch_json_object(url):
            '''Naively try to fetch a JSON result from the given url and return
            a parsed python object. Let the FuncWrap handle any exceptions.'''
            res = urllib2.urlopen(url)
            return json.loads(res.read())
        
        # validation and initialisation is done inline here.
        if not queries:
            raise ValueError('Must provide at least one pending query! Got none!')
        _validate_type('num_worker_threads', num_worker_threads, types.IntType) 
        _validate_type('queries', queries, types.DictType)
        
        self.func_wraps = []
        for q_name, q_url in queries.iteritems():
            try:
                _validate_type('', q_name, types.StringType)
                _validate_type('', q_url, types.StringType)
            except TypeError:
                raise TypeError('queries must be a dict of "str_query_name/id": "str_url_to_fetch". ' + 
                    'Instead, got %s: %s in %s' %(q_name, q_url, queries))

            self.func_wraps.append(
                thr.FuncWrap(target=fetch_json_object, name=q_name, args=(q_url,))
            )
        
        self.pool = thr.FuncWrapPool(num_worker_threads, self.func_wraps)
        self.result = {}
        # skipping validation for now, will add this code later

    @property
    def is_finished(self):
        return bool(self.result)

    @property
    def has_exception(self):
        if not self.result:
            return False
        return any(isinstance(r, Exception) for r in self.result.values())

    def run(self):
        '''Run the queries in the FuncWrapPool, await results and then collect them in
        self.result in the form { 'query_str_label': result_obj } where the result_obj
        is either a parsed python result if it worked or an exception if it didn't!'''
        self.pool.run()
        for fw in self.func_wraps:
            if fw.exception:
                self.result[fw.name] = fw.exception
            else:
                self.result[fw.name] = fw.result
            



__author__ = 'nickmab'

'''Utility code for use by other py files within this package.'''

def validate_type(obj_name, obj, desired_type):
    if not isinstance(obj, desired_type):
        raise TypeError('%s must be a %s. Instead got %s: %s.'
            %(obj_name, desired_type.__name__, type(obj).__name__), obj)
